#!/bin/sh
# the next line restarts using tcl/tk interpreter\
exec wish "$0" "$@"

#
#  SUPERVISOR: Observation process manager
#
namespace eval sv {

# The only hardcoded setting is a name of configuration file:
variable cfg_file_name [pwd]/sv.cfg
# version:
variable version 0.01

# Components list:
variable comp_list [list]
# 
#... Reading the configuration file
#
proc read_cfg file {
	set in [open $file r]
	# Until first "component xxx" line the parameters are those of SV:
	set comp SV
	variable ::SV
	set linenum 0
	while {[gets $in line] >= 0} {
		incr linenum
		# Locate possible comment character:
		if {[set ci [string first \# $line]] >=0} { 
			# remove righthand comment after ci:
			set line [string range $line 0 [expr $ci-1]]
		}
		# delete possible leading/trailing spaces:
		set line [string trim $line]	
		# locate parameter name end:
		set dp [string wordend $line 0]
		set sl [string length $line]
		if {$dp<$sl} {
			set pname [string tolower [string range $line 0 [expr $dp-1]]]
			set pvalue [string trim [string range $line \
				[string wordstart $line	$dp] end]]
			if {$pname=="component"} {
				# declaration of the next component section
				set comp $pvalue
				# new component or addition to previously defined?:
				if {[lsearch $sv::comp_list $comp]<0} {
					# add new component to the list of known components
					lappend sv::comp_list $pvalue
					# create the global array of parameters:
					variable ::$pvalue
					# create the same name procedure for accessing array
					set s "{ return \$::$pvalue\(\$param\) }"
					uplevel #0 proc $pvalue param $s
				}
			} elseif {$pname=="alias"} {
				# declaration that the current component is an alias of another
				# delete this component
				if {[info exists ::$comp]} {unset ::$comp}
				# delete this component from the list as well, since no 
				# personal init/stop/park is needed for it:
				set i [lsearch $sv::comp_list $comp]
				set sv::comp_list [lreplace $sv::comp_list $i $i]
				upvar #0 ::$pvalue ::$comp
			} else {
				set ${comp}($pname) $pvalue
			}
		} else {
			if {$dp>0} {error "Haven't got line $linenum in $file"}
		}
	}
	close $in
}

#-------------------------- COMMAND SERVICE --------------------------------
# Array of lists of "after" results and commands, for which these
# after-timers were set, indexed by cmd_id,comp pairs: e.g. 
# cmd_list(0,OBJM) = {after#1 {GET ID}}
# After receiving a reply, the element must be unset.
variable cmd_list

# Common sv-namespace variable - command identifier
variable cmd_id 0

#
# Sending a command to a component. Return command identifier on success, -1 on error
#
# For displaying the commands history tail:
variable answered_cmd [list]
variable answered_cmd_n 4

proc cmd {comp kwd args} {
	global $comp SV
	upvar #0 $comp cmp
	set kwd [string toupper $kwd]
	
	# Send a command string to the component and register and schedule its 
	# timeout:
	set chanid $cmp(chanid)
	# - chanid here is assigned in connect()

	# Detect request for background execution command: no waiting for completion, 
	# only for the first reply:
	
	if {[set i [lsearch $args "\&"]]>=0} {
		set args [lreplace $args $i $i]
		set bkgrexec 1
	} else {
		set bkgrexec 0
	}

	set cmd "$kwd $args"
	if {$chanid=="closed"} {
		# component was disconnected by closing the channel
		error "cannot send commands to disconnected component $comp"
	} elseif {$chanid=="disabled"} {
		# component was disabled due to its error state
		return -1
	} else {
		puts $chanid "$sv::cmd_id $cmd"
		flush $chanid

		# Register the sent command in an array
		set sv::cmd_list($sv::cmd_id,$comp) [list \
			[after $::SV(tmout) {error "Command lost"}] $cmd]

#puts "send: cmd_list($comp)=[array get sv::cmd_list *,$comp]"
	}
	set new_cmd "->$comp $sv::cmd_id $cmd"
	add_log $new_cmd
	set prev_cmd_id $sv::cmd_id
	# Modify the command identifier according to the 0..65535,0.. cycling rule:
	if {$sv::cmd_id==65535} { set sv::cmd_id 0 } else { incr sv::cmd_id }

	# Delay the return from the command until it's replied:
	# Wait until variable is unset. It may be set again by WAIT-prolongation.
	while {[info exists sv::cmd_list($prev_cmd_id,$comp)]} {
			vwait sv::cmd_list($prev_cmd_id,$comp) 
			if {$bkgrexec} break
	}
	return $prev_cmd_id
}

#
# Reply whether the command with given $comid is still under execution (1/0)
#
proc is_cmd comid {
	if {[llength [array get sv::cmd_list $comid,*]]>0} {return 1} {return 0}
}

#
# Wait until command execution is finished. Returns 0
#
proc wait_cmd comid {
	while {[is_cmd $comid]} {vwait sv::cmd_list}
	return 0
}

source parsecmd.tcl

proc get_reply {comp args} {
	global $comp
	upvar #0 $comp cmp
	if {[gets $cmp(chanid) reply] < 0} {
		# end of file condition: disactivate current handler
		close $cmp(chanid)
		error "Cannot read from $comp channel"
		return
	}
	add_log "<-$comp $reply"
	
	set rl [parse_command $reply]

	# Extract command identifier and keyword:
	set ci [ lindex $rl 0 ]
	set kwd [lindex $rl 1]
	# Remove COMID and KWD from reply leaving the parameter=value's only:
	set rl [lreplace $rl 0 1]
#puts "receive1: cmd_list($comp)=[array get sv::cmd_list *,$comp]"
#puts "After=[lindex $sv::cmd_list($ci,$comp) 0]"
	# Cancel the timeout timer for this command and delete it from list:
	set cmd [lindex $sv::cmd_list($ci,$comp) 1]
	if {[catch {after cancel [lindex $sv::cmd_list($ci,$comp) 0]}]>0} { 
		# if caugth then no timer for such COMID was registered for COMP
		error "Unexpected reply $reply arrived from $comp"
	}
	
	# Parse rest of reply and assign the replied quantities to the component 
	# parameters:
	
	if {[lsearch $rl switch]>=0} {
		error "Non-conventional reply $reply from $comp"
	} else {
		# remove wait-parameter if it was a delayed result 
		array unset $comp wait
		array set $comp $rl
		if {[llength [array names $comp wait]]>0} {
			# Postpone final reply getting a bit more:
			set wms [expr $cmp(wait) * 1000 + $::SV(tmout)]
#puts "Postpone reply timeout by another $wms ms"
			# Register another timer of the command in an array
			set sv::cmd_list($ci,$comp) [list \
				[after $wms {error "Command lost"}] \
				[lindex $sv::cmd_list($ci,$comp) 1] ]
#puts "postpone: cmd_list($comp)=[array get sv::cmd_list *,$comp]"
		}	
	}
	# Remove the command from the array if there was no wms set (WAIT parsed).
	# We could not do it before parsing since there must be no period of
	# respective element ansence (vwait in cmd()).
	if {[info exists wms]==0} {
		# append the current answered command to the list and remove oldest:
		lappend sv::answered_cmd "$ci ->$comp $cmd  <- $reply"
		if {[llength $sv::answered_cmd]>$sv::answered_cmd_n} {
			set sv::answered_cmd [lreplace $sv::answered_cmd 0 0]
		}
		array unset sv::cmd_list $ci,$comp
	}
	
	switch $kwd {
		"ERROR"
			{ 	
				# if error was caused by BUSY status, repeat when status changed
				if {$cmp(status)=="BUSY"} {
					add_log "wait while $comp is $cmp(status)..."
					vwait ::${comp}(status)
					add_log "Ok: $comp is $cmp(status): do $cmd to $comp"
					cmd $comp $cmd
				} else {
					add_log "Error answer $reply from $comp. Ignored" 
				}
			}
		"OK"
			{ }
		default
			{ error "Error keyword in reply $reply from $comp" }
	}

#puts "receive: cmd_list($comp)=[array get sv::cmd_list *,$comp]"
#puts "$comp params:[array get ::$comp]"
}


#------------------------- TIME SERVICE ------------------------------------
#
#... Return the local time stamp according to the adopted format:
#
proc time_stamp {{sec1969 -1}} {
	if {$sec1969==-1} {
		set sec1969 [clock seconds]
	}
	return [clock format $sec1969 -format "%Y-%m-%d %H:%M:%S" ]
}

#------------------------ ESTABLISH COMMUNICATION ---------------------------

#
#... Connect all components mentioned in CFG
#
proc connect args {
	if {$args=="all"} {set args $sv::comp_list}
	foreach comp $args {
		global $comp 
		upvar #0 $comp cmp
		add_log "Connect $comp at $cmp(host) port $cmp(port)"
		# Check if comp is started:

		if {[catch {set t [socket $cmp(host) $cmp(port)]} ]} {
			error "Component $comp is not accessible"
        } else {
			set cmp(chanid) $t
			# Create the procedure for handling the replies (Overwrites existing):
			uplevel #0 "proc get_reply_$comp {} { sv::get_reply $comp }"
			# Set it as a component channel reading handler:
			fileevent $cmp(chanid) readable get_reply_$comp
        }
		
		set ident_cfg $cmp(ident)
		set c1 [cmd $comp GET IDENT]
		if {$cmp(ident)!=$ident_cfg} {
			disconnect $comp
			error "Component $comp is $cmp(ident) while should be $ident_cfg"
		}
		cmd $comp GET STATUS
		# Park if not parked?
	}
}

#
#... Disconnect all components (before quitting)
#
proc disconnect args {
	if {$args=="all"} {set args $sv::comp_list}
	foreach comp $args { 
		global $comp
		upvar #0 $comp cmp
		add_log "Disconnect $comp"
		# the channel reading handler will be unset automatially now:
		close $cmp(chanid)
		set cmp(chanid) closed
	}
}

#------------------------ PARKING/INITIALIZATION ---------------------------
#
#...Initialize components
#
proc initialize complist {
	if {$complist=="all"} {
		set clist $sv::comp_list
	} else {
		upvar $complist clist
	}
	# Issue the Initialize command to all components:
	foreach comp $clist { 
		lappend initcmdlist [cmd $comp INIT &]
	}
	# wait for all init commands are over:
	add_log "Wait for components to finish initialization..."
	foreach c $initcmdlist {
		wait_cmd $c
	}
	add_log "Initialization finished."	
}

#
#...Stop and park components
#
proc stop_park complist {
	if {$complist=="all"} {
		set clist $sv::comp_list
	} else { 
		upvar $complist clist
	}
	foreach comp $clist {
		cmd $comp STOP NOW
		lappend parkcmdlist [cmd $comp PARK &]
	}
	# wait for all park commands are over:
	add_log "Wait for components to finish parking..."
	foreach c $parkcmdlist {
		wait_cmd $c
	}
	add_log "Parking finished."

}

proc shutdown {} {
	stop_park all
	disconnect all
	add_log "bye."
}

#------------------------ SCENARIA MANAGEMENT -------------------------------
#
#...Read and check the observation scenario
#
proc check_scen args {
	foreach s $args { 
		puts -nonewline "Check scenario $s..."
		if {[file readable $s]} {
			puts "Ok"
		} else { puts "Bad"; error "Scenario $s is not readable" }
	}
}

# Flag of observations going on: global
variable is_observations_now 0

# Text representation of is_observations_now:
variable obsmon_state "Monitoring Conditions"

trace variable sv::is_observations_now w sv::swap_obs_mon

proc swap_obs_mon {name1 name2 op} {
	if {$sv::is_observations_now} {
		set sv::obsmon_state "Observations"
	} else { set sv::obsmon_state "Monitoring Conditions" }
}

#
#... Return is_observations_now
#
proc is_observations_now {} {return $sv::is_observations_now}

#
#...Update idle callbacks for refreshing the SV window
#
proc update_idle delay { 
	update idletasks; 
	if {$sv::is_observations_now} {
		after $delay sv::update_idle $delay 
	}
}

#
#...Start observations scenario
#
# Tcl interpreter for running observation scenario
variable tclobs 0
# flag of cancellation of observations BEFORE SV returns:
variable disable_start_obs false

proc start_obs {} {
	global SV tclobs
	# detect whether the stop_obs was forced before with disable_start_obs==true
	if {$sv::disable_start_obs} return
	# determine which components are not initialized (by their status)
	# (defined at least in connect()). Circ.monitor MUST initialize components
	# which it needs! - namely METEO and CIRC.
	set sv::tclobs [interp create -safe]
	set sv::is_observations_now 1
	sv::initialize SV(obs_comp_list)

	# make aliases for accessing commands in the interpreter running monitor:
	if {$sv::is_observations_now} {
		sv::add_log "Start observations: interpreter $sv::tclobs"
		foreach c {cmd is_cmd wait_cmd} {$sv::tclobs alias $c $c}		
		foreach c $sv::comp_list { $sv::tclobs alias $c $c}
	# below catch catches the eval of deleted interp. commands when stopped obs.
		after 10 {
			add_log "Source $SV(oscen)"
			catch {$sv::tclobs invokehidden -global source $SV(oscen)}
		}
	}
	update_idle 2000
}

#
#...Stop observations 
#
proc stop_obs {{disable_start_obs false}} {
	global SV
	if {!$sv::is_observations_now} return
	set sv::disable_start_obs $disable_start_obs
	sv::add_log "Stop observations: stop interpreter $sv::tclobs"	
	set sv::is_observations_now	0
	interp delete $sv::tclobs
	sv::stop_park SV(obs_comp_list)
}

#
#...Start monitor
#
variable tclmon 0
proc start_monitor monitname {
	global SV tclmon
	add_log "Start monitor $monitname..."
	set tclmon [interp create -safe]
	# make aliases for accessing commands in the interpreter running monitor:
	foreach c {cmd is_cmd wait_cmd start_obs stop_obs add_log is_observations_now} {
		$tclmon alias $c $c
	}
	foreach c $sv::comp_list { $tclmon alias $c $c}
	after 100 $tclmon invokehidden -global source $monitname
}

#-------------------------- INTERFACE, LOGGING ------------------------------
proc add_log text {
	puts "[time_stamp] $text"
}

proc create_window {} {
	package require Tix
	wm withdraw .
	set w [toplevel .t]
	wm title .t "Supervisor v$sv::version"
	wm protocol .t WM_DELETE_WINDOW sv::terminate
	
	# create frame-table with sub-frames-columns each for its component items
	set f [frame $w.ft]
	pack [label $f.tlabel -text Components] -side top
	pack $f -fill x -padx 2 -pady 2
	
	set f [frame $w.fcmp -borderwidth 3]
	foreach c $sv::comp_list { 
		global $c
		upvar #0 $c cmp
		set fc [frame $f.f$c -bd 3 -width 100 -relief groove]
		# button with component name:
		menubutton $fc.cn -text $c -state normal -menu $fc.cn.menu
		set m [menu $fc.cn.menu]
#		set cm  "set ${c}(status) ISOLATED"
#		$m add command -label Isolate -command $cm
#		set cm "cmd $c get status"
#		$m add command -label Activate -command $cm
		$m add command -label Command -command "sv::cmd_dialog $c"
		label $fc.status -textvariable ${c}(status) -fg red
		pack $fc.cn $fc.status -fill x -padx 2 -pady 2 -side top
		if {[info exists cmp(show)]} {
			foreach v $cmp(show) {
			if {$v=="data"||$v=="status"} continue
				frame $fc.v$v
				pack [label $fc.v$v.lab -text "$v:" ] -side left
				pack [label $fc.v$v.val -textvariable ${c}($v) -fg blue] -side \
					left
				pack $fc.v$v -anchor w
			}
		}
		pack $fc -side left -anchor n -expand true -fill y
	}
	pack $f -pady 2 -expand true
	
	# Create a section with components data lines:
	set f [frame $w.fdata]
	foreach c $sv::comp_list {
		global $c
		upvar #0 $c cmp
		if {[info exists cmp(show)]==0 || [lsearch $cmp(show) data]<0} continue
		set fc [frame $f.f$c]
		pack [label $fc.dlabel$c -text "$c DATA:"] -side left -anchor w
		pack [label $fc.data$c -textvariable ${c}(data) -fg blue] -side left -anchor w
		pack $fc -padx 2 -pady 2 -anchor w
	}
	pack $f -pady 2 -anchor w

	# Create the message with observations/monitoring state:
	set f [frame $w.obs]
	pack [label $f.lab -text "Activity:" ] -side left
	pack [label $f.val -textvariable sv::obsmon_state -fg blue] -side \
			left
	pack $f -pady 2
	
	
	# Create the window with the last and delayed-execution commands:
	set f [frame $w.cmd -takefocus 0]
	pack [label $f.label -text "Last commands:"]
	pack [listbox $f.cmdlist -takefocus 0 -selectmode none -height \
		[expr [llength $sv::comp_list]+$sv::answered_cmd_n]  -width 60]
	pack $f -padx 2 -pady 2 -anchor w
	trace variable sv::cmd_list wu sv::update_cmd_list
	
	# Commands to manipulate observations state:
	set f [frame $w.obscmd]
	pack [button $f.stop -text "Stop Observations" -state disabled \
		-command {sv::stop_obs true}] -side left -padx 10
	pack [button $f.enable -text "Allow Observations" \
		-command {set sv::disable_start_obs false}]  -side left -padx 10
	pack [button $f.terminate -text "Terminate" \
		-command sv::terminate]  -side left -padx 10
	pack $f -pady 2
	trace variable sv::is_observations_now w sv::update_obs_buttons
}

#
#...Depending on is_observations_now, activate or disable buttons:
#
proc update_obs_buttons {name1 name2 op} {
	if {$sv::is_observations_now} {
		.t.obscmd.stop configure -state active
	} else {
		.t.obscmd.stop configure -state disabled	
	}
}

#
#...Operator's dialog with a component from Tk window:
#
variable tix_combo_box {}
variable label_cont ""
proc cmd_dialog comp {
	toplevel .w
	wm title .w "SV command dialog"
	set sv::label_cont "Enter command to $comp"
	set f [tixComboBox .w.cmd -selectmode browse -editable true -dropdown true\
		-variable sv::tix_combo_box -padx 20 -pady 4\
		-labelside top -fancy true -history true -historylimit 20 \
		-command "sv::cmd_issue $comp" ]
	.w.cmd subwidget label configure -width 100 -textvariable sv::label_cont
	foreach cmnd { INIT RUN STOP PARK FREE GET SET } { .w.cmd insert end $cmnd } 
	pack .w.cmd
}
proc cmd_issue {comp cmd} {
	cmd $comp $cmd
	set reply [lindex $sv::answered_cmd end]
	set sv::label_cont "Reply is: $reply\n New command:"
}

#
#...Update the view of the commands list in the SV window: invoked by trace var.
#... First creates the new list of delayed commands, then deletes the old
#...displayed list and adds the sorted concatenation of the delayed commands and
#...answered commands
#
proc update_cmd_list {name1 name2 op} {
	set cl [list]
	foreach n [array names sv::cmd_list] {
		set s "$n [lindex $sv::cmd_list($n) 1]"
		set s [string replace $s [set i [string first , $s]] $i " ->"]
		lappend cl $s
	}
	set cls [lsort -dictionary [concat $cl $sv::answered_cmd]]
	.t.cmd.cmdlist delete 0 [expr [.t.cmd.cmdlist size] - 1]
	foreach e $cls {
		.t.cmd.cmdlist insert end $e}
}

proc terminate {} {
		if {$sv::is_observations_now} {stop_obs true}
		trace vdelete sv::cmd_list wu sv::update_cmd_list
		trace vdelete sv::is_observations_now w sv::update_obs_buttons
		destroy .t
}

###################### Starting supervisor:

proc main {} {
	global SV
	read_cfg $sv::cfg_file_name
	# listing the parameters of SV and components and making it global in main()
#	foreach c $sv::comp_list { global $c; puts "Component $c:"; puts [$c port] }
	create_window
	connect all
	check_scen $SV(cscen) $SV(oscen)
	# arrival of TERM signal will remove the SV window causing tkwait...t to end
#	signal trap SIGTERM { sv::terminate } 
	start_monitor $SV(cscen)
	tkwait window .t
	shutdown
}

namespace export -clear cmd is_cmd wait_cmd start_obs stop_obs add_log is_observations_now

}

# import those commands of namespace sv which are used in scenaria:
namespace import sv::cmd sv::is_cmd sv::wait_cmd sv::start_obs sv::stop_obs sv::add_log sv::is_observations_now
sv::main
exit
