#!/bin/sh
# the next line restarts using tclsh\
exec wish "$0" "$@"

#
#  SUPERVISOR: Observation process manager
#

#
# Global settings and  variables:
#
set sv_dir [ pwd ]
set sv_cfg_file ${sv_dir}/sv.cfg

# Main global parameters read from CFG file and (maybe) further modified
# Status cases: unknown, free, busy, ready, error
# Statuses busy and ready refer to TIEd status, free to FREEd, error - any,
# unknown - no connection yet
set sv_p(DET,status) unknown
set sv_p(TLSP,status) unknown
set sv_p(OBJM,status) unknown
set sv_p(METEO,status) unknown
# Component channel identifiers initialization:
set sv_p(DET,chanid) unknown
set sv_p(TLSP,chanid) unknown
set sv_p(OBJM,chanid) unknown
set sv_p(METEO,chanid) unknown

# Night reference times in secs after 1969: "-1" means undefined
# Beginning of restricted mode observations: end of civil twilight

# DEBUG: set night 70 seconds after now given max preptime 60, 10s twilights and 20s night
set t1 [expr [clock seconds] + 70]
# End of astronomical twilight: beginning of "unlimited" obs-s
set t2 [expr $t1 + 10] 
# beginning of morning restricted obs-s: beginning of astronomical twilight:
set t3 [expr $t2 + 20]
# end of observations: beginning of morning civil twilight
set t4 [expr $t3 + 10]
# Note: when t2=t3 returned by OBJM, this means that astronomical evening
# twilight continues by morning twilight; t2=t3=(t1+t4)/2;
# when polar day/white nights is on the calendar; t1..t4 are returned for
# the first night after summer solstice.
# The rule is that t4 must be more than the current time, i.e. if morning 
# twilight is over, then t1-t4 are computed for the next night.

# Commands counter for COMID in exchange protocol:
set comid 0
set sv_name $argv0

# Actual meteo data: humidity %, temperature degC, wind speed m/s, direction
# azimuth, clouds %
set meteo(humid) 0
set meteo(temp) 0
set meteo(wind,speed) 0
set meteo(wind,dir) 0
set meteo(cloud) 0

# 
#... Reading the configuration file:
#
proc cfg_read cfg_file {
	global sv_p sv_name
	if {[file exist $cfg_file] == 1 && [file readable $cfg_file] == 1} {
		set in [open $cfg_file r]
		while {[gets $in line] >= 0} {	
			# Locate possible comment character:
			if {[set ci [string first \# $line]] >=0} { 
				# remove righthand comment after ci:
				set line [string range $line 0 [expr $ci-1]]
			}
			if {[string first = $line]>0} {
				set sv_p([lindex [split $line =] 0]) \
				[eval concat [lreplace [split $line =] 0 0]]
			}
		}
		close $in
	} else {
		sv_fatal "Configuration file $cfg_file open error!"
	}
#	foreach par [array names sv_p *preptime,max] {
#		puts "sv_p($par)=$sv_p($par)"
#	}
}

proc sv_start {} {
	global sv_cfg_file sv_p t1 t2 t3 t4
	cfg_read $sv_cfg_file
	
	sv_open_comp DET
	
	sv_open_comp OBJM

	sv_command OBJM get t1 t2 t3 t4
	# Check that t4 is not in the past:
	if {[sv_wt $t4]<=0} {
		# Maybe an exact coincidence of current time and t4 - wait a bit:
		after 5000
		sv_command OBJM get t1 t2 t3 t4
		if {[sv_wt $t4]<=0} {
			sv_fatal "OBJM error: reports on T4 in the past. Stop"
		}
	}
	# Create a list ptimes of maximal preparation times :
	foreach par [array names sv_p *preptime,max] {lappend ptimes $sv_p($par)}
	# Derive the beginning time from maximal of ptimes:
	set begtime [expr $t1 - [lindex [lsort -real -decreasing $ptimes] 0]]

	if { [set dt [sv_wt $begtime]] > [expr $sv_p(maxobjmwait)*1000] } {
		sv_log "Postpone preparations to observations at $dt ms..."
		sv_close_comp OBJM
	}
	# recalculate delay here since sv_close_comp OBJM might took some time...
	after [sv_wt $begtime] sv_startprep
}

#
# Start and connect to the component. Return connected flag
#
proc sv_open_comp comp {
	global sv_p comreply

	# Check if comp is connected: comp might not reply!
	if {$sv_p($comp,chanid)!="unknown" && \
		![catch {sv_command $comp get status} fid]} {
		set sv_p($comp,status) $comreply($comp,STATUS)
		return true  
	}

	# evidently not connected
	sv_log "Start component $comp"

	# Check if comp is started:.................
	if {[catch {set t [socket $sv_p($comp,host) $sv_p($comp,port)]} fid]} {
		sv_fatal "Component $comp not started or has a wrong address"
	} else {
		set sv_p($comp,chanid) $t
	}
	
	# Create socket connection to server-like component:
	sv_log "Component $comp got channel $sv_p($comp,chanid)"

puts "Send get ident:"
	sv_command $comp init
	sv_command $comp GET IDENT
puts "get id sent"
	
	if {$comreply($comp,IDENT)!=$comp} {sv_fatal "Bad $comp address!"}

	sv_command $comp get status
	set sv_p($comp,status) $comreply($comp,STATUS)
}

proc sv_close_comp comp {
	sv_log "Stop component $comp"
	if {([info exists sv_p($comp,quit)]>0) && $sv_p($comp,quit)} {
		sv_command $comp park quit			
	} else {
		sv_command $comp park
	}
	close $sv_p($comp,chanid)
}

proc sv_startprep {} {
	global t1 t2 t3 t4 sv_p
	sv_log "Schedule components start..."
	foreach comp [list METEO TLSP DET OBJM] {
		# delay starting comp preparation at
		after [set dt [sv_wt [expr $t1-$sv_p($comp,preptime,max)]]] \
			sv_open_comp $comp
		if {$dt>0} {sv_log "Delay $comp start at $dt ms"}
	}

	sv_log "Schedule night and twilights reftimes..."

	after [set dt [sv_wt $t1]] {sv_set_twilight true; sv_start_observations}
	if {$dt>0} {sv_log "Schedule obs. beginning at $dt ms"}

	after [set dt [sv_wt $t2]] {sv_set_twilight false}
	if {$dt>0} {sv_log "Schedule evening end at $dt ms"}

	after [set dt [sv_wt $t3]] {sv_set_twilight true}
	if {$dt>0} {sv_log "Schedule morning begin at $dt ms"}

	after [set dt [sv_wt $t4]] {sv_end_observations; sv_set_twilight false}
	if {$dt>0} {sv_log "Schedule obs. end at $dt ms"}
		
}

#
# ... Compute the time interval between the specified moment in future 
# (given in seconds after 1969) and the current moment and return it in ms
# If given moment is in the past, return 0
#
proc sv_wt sec {
	set d [expr int ($sec - [clock seconds])]
	if {$d>0} {return [expr $d*1000]} else {return 0}
}

# Array of lists of "after" results and commands, for which these
# after-timers were set, indexed by comp,comid pairs: e.g. 
# comlist(OBJM,0) = {after#1 {GET ID}}
# After receiving a reply, the element must be unset.
set comlist(OBJM,0) [list]

# Send a command string to the component and register and schedule its timeout
proc sv_comsendr {comp comid comm nsnd} {
	global sv_p comlist
	# Attention! comid here is LOCAL, since the call may be a command repetition
	set chanid $sv_p($comp,chanid)

puts "$comid $comm"	
	puts $chanid "$comid $comm \n"
	flush $chanid

	# Register the command sent in an array
	set comlist($comp,$comid) [list \
		[after $sv_p(timeout) {sv_command_lost}] $comm $nsnd]
}

#
# Creation and sending a command to a component. Return 0 if success, error code otherwise
#
proc sv_command {comp kwd {args no}} {
	global comid sv_p comlist
	set key [string toupper $kwd]
	sv_log "COMMAND \#$comid: $key to $comp with arguments $args"
	
	# send and register the command with nsnd=1:
	sv_comsendr $comp $comid "$key $args" 1

	puts "comlist($comp)=[array get comlist $comp,*]"

	fileevent $sv_p($comp,chanid) readable sv_get_reply_$comp
	
	incr comid
	# Delay the return from the command until it's replied if the result is needed: key=GET:
	if {$key=="GET"} { 
		# Save command identifier for current command to given component:
		set comid2($comp) [expr $comid - 1 ]
		
		# Wait until variable is unset. It may be set again by command repetition.
		while {[info exists comlist($comp,$comid2($comp))]} {
			vwait comlist($comp,$comid2($comp)) 
		}
	}
	return true
}

# Handler of unanswered commands:
proc sv_command_lost {} {
	global comlist sv_p
	sv_log "COMMAND LOST"
#	sv_fatal "Component doesn't reply"
	# create a list of postponed tasks:
	set afl [after info]
	set cml [array startsearch comlist]
	while {[set eln [array nextelement comlist $cml]]>""} {
		if {[ lsearch -exact $afl [lindex $comlist($eln) 0] ] < 0} {
			# found afterID not listed in after info list. 
			# It's the cause of the current call.
			set compname [lindex [split $eln ,] 0] 
			set ci [lindex [split $eln ,] 1]
			set comm [lindex $comlist($eln) 1]
			set nsnd [lindex $comlist($eln) 2]
puts "Lost: \# $ci $comm to $compname sent $nsnd times"
			if { $nsnd < $sv_p(maxnsnd) } {
				# resend the command:
				sv_comsendr $compname $ci $comm [incr nsnd]
			} else {
				# all attempts are over:
				sv_restart_comp $compname 
			}
		}
	}
}

proc sv_restart_comp comp {
	sv_fatal "Restart component $comp"
}

proc sv_unexpected_reply {reply comp} {
	sv_fatal "Unexpected reply $reply arrived from $comp"
}

# Individual file event handlers assigned in sv_command:
proc sv_get_reply_OBJM {} {sv_get_reply OBJM}
proc sv_get_reply_DET {} {sv_get_reply DET}
proc sv_get_reply_TLSP {} {sv_get_reply TLSP}
proc sv_get_reply_METEO {} {sv_get_reply METEO}

# Global array of replied parameters from components:
# both component names and parameter names are put here uppercase by get_reply:
set comreply(OBJM,ID) 0

proc sv_get_reply comp {
	global sv_p comlist comreply
	if {[gets $sv_p($comp,chanid) reply] < 0} {
		# end of file condition: disactivate current handler
		fileevent $sv_p($comp,chanid) readable ""
		return
	}
	# Extract command identifier and keyword and remove them from list:
	set ci [ lindex [ set rl [ split $reply ] ] 0 ]
	set kwd [string toupper [lindex $rl 1]]
	set rl [lreplace $rl 0 1]
	
	puts "REPLY from $comp is: $reply, comid=$ci kwd=$kwd"
	puts "comlist($comp)=[array get comlist $comp,*]"

	# Cancel the timeout timer for this command and delete it from list: 
	if {[catch {after cancel [lindex $comlist($comp,$ci) 0]}]>0} { 
		sv_unexpected_reply $reply $comp
	}
	
	# Check keyword:
	if {$kwd == "OK"} {
		# Reply is positive. Parse it:
		foreach pv $rl {
			# convert pv="p=v" into pv={p v}
			set pv [split $pv =]
			
			# Whether argument is WAIT=...?
			if {[lindex $pv 0] == "WAIT"} {
				# Postpone final reply getting a bit more:
				set wms [expr [lindex $pv 1] * 1000 + $sv_p(timeout)]
				puts "Postpone reply timeout by another $wms ms"
				# Register the command sent in an array
				set comlist($comp,$ci) [list \
					[after $wms {sv_command_lost}] \
					[lindex $comlist($comp,$ci) 1] \
					[lindex $comlist($comp,$ci) 2]	]
			} else {
				# Save globally replied argument names and their values:
				set comreply($comp,[string toupper [lindex $pv 0]]) \
					[lindex $pv 1]
			}
		}
	} else {
		if {$kwd == "ERROR"} {
			sv_comp_error $comp
		} else {
			sv_log "Unrecognized reply from $comp with key=$kwd"
		}
	}
	# delete command info if no WAIT=* came and thus unblock sv_command with GET
	if {[info exists wms] == 0} {unset comlist($comp,$ci)}

	puts "comlist($comp)=[array get comlist $comp,*]"
	puts "comreply($comp)=[array get comreply $comp,*]"
}

# Reaction on the error state replied by the component
proc sv_comp_error comp {
	sv_fatal "Error state in component $comp"
}

proc sv_start_observations {} {
	sv_log "Start observations"
	# Wait for all the components to be ready:
}

proc sv_set_twilight istwi {
	if {$istwi} {
		sv_log "Twilight set"
	} else {
		sv_log "Twilight reset"
	}
	set twiflag $istwi
}


proc sv_end_observations {} {
	global sv_p
	sv_log "End observations"

	foreach comp [list DET TLSP METEO] {
		
		# Wait for component to finish:
		sv_command $comp stop
		sv_close_comp $comp
	}

	# Schedule the tomorrow preparations:
	sv_start
}

###################### GUI and I/O

proc sv_message msg {
	global sv_name
	tk_dialog .d "$sv_name action" $msg info 0 OK
}


proc sv_log msg {
	puts $msg
}

proc sv_fatal msg {
	global sv_name
	bell
	tk_dialog .d "$sv_name fatal error" $msg error 0 OK
	exit
}

proc sv_create_window {} {
	wm withdraw .
	set w [toplevel .t]
	wm title .t {Supervisor}
}

###################### Starting supervisor:

proc main {} {
	sv_create_window
	sv_start
	tkwait window .t
}

main
